"""
Account registration.
"""

import logging
import os

from .account import Account
from .acme import Acme
from .errors import ManualeError, AccountAlreadyExistsError
from .crypto import generate_rsa_key
from .helpers import confirm

logger = logging.getLogger(__name__)

def register(server, account_path, email):
    # Don't overwrite silently
    if os.path.exists(account_path):
        if not confirm("The account file {} already exists. Continuing will overwrite it with a new key. Continue?".format(account_path), default=False):
            raise ManualeError("Aborting.")

    # Confirm e-mail
    if not confirm("You're about to register a new account with the e-mail {}. Continue?".format(email)):
        raise ManualeError("Aborting.")

    # Generate the account key
    logger.info("Generating a new account key. This might take a second.")
    account = Account(key=generate_rsa_key(4096))
    logger.info("Key generated.")

    # Register
    acme = Acme(server, account)
    logger.info("Registering...")
    try:
        registration = acme.register(email)

        # If the server has terms of service, prompt the user to confirm
        # TODO: This is a really stupid flow. Keep an eye out on this issue:
        # https://github.com/ietf-wg-acme/acme/issues/59 and hope they fix it.
        if registration.terms:
            logger.info("This server requires you to agree to these terms:")
            logger.info("  {}".format(registration.terms))
            if not confirm("Agreed?"):
                logger.error("Aborting. Your account was still created, but it won't be usable before agreeing to terms.")
                raise ManualeError()
            acme.update_registration({ 'agreement': registration.terms })
            logger.info("Updated account with agreement.")

        logger.info("Account {} created.".format(account.uri))
    except IOError as e:
        logger.error("Registration failed due to a connection or request error.")
        raise ManualeError(e)

    # Write account
    directory = os.path.dirname(os.path.abspath(account_path))
    os.makedirs(directory, exist_ok=True)
    with open(account_path, 'wb') as f:
        f.write(account.serialize())

    logger.info("Wrote account to {}.".format(account_path))
    logger.info("")
    logger.info("What next? Verify your domains with 'authorize' and use 'issue' to get new certificates.")



"""
ACME API client.
"""

import copy
from collections import namedtuple
from urllib.parse import urljoin, urlparse

import requests

from . import __version__
from .crypto import generate_header, sign_request
from .errors import *

DEFAULT_HEADERS = {
    'User-Agent': "manuale {} (https://github.com/veeti/manuale)".format(__version__),
}

class Acme:

    def __init__(self, url, account):
        self.url = url
        self.account = account
        self.key = account.key

    def get_nonce(self):
        """
        Gets a new nonce.
        """
        return self.get('/directory').headers.get('Replay-Nonce')

    def get_headers(self):
        """
        Builds a new pair of headers for signed requests.
        """
        header = generate_header(self.key)
        protected_header = copy.deepcopy(header)
        protected_header['nonce'] = self.get_nonce()
        return header, protected_header

    def register(self, email):
        """
        Registers the current account on the server.
        """
        response = self.post('/acme/new-reg', {
            'resource': 'new-reg',
            'contact': [
                "mailto:{}".format(email)
            ],
        })
        uri = response.headers.get('Location')
        if response.status_code == 201:
            self.account.uri = uri

            # Find terms of service from link headers
            terms = _get_link(response.headers.get('Link'), 'terms-of-service')

            return RegistrationResult(
                contents=_json(response),
                uri=uri,
                terms=terms
            )
        elif response.status_code == 409:
            raise AccountAlreadyExistsError(response, uri)
        raise AcmeError(response)

    def get_registration(self):
        """
        Gets available account information from the server.
        """
        response = self.post(self.account.uri, {
            'resource': 'reg',
        })
        if str(response.status_code).startswith('2'):
            return _json(response)
        raise AcmeError(response)

    def update_registration(self, params=None):
        """
        Updates registration information on the server.
        """
        params = params or {}
        params['resource'] = 'reg'

        response = self.post(self.account.uri, params)
        if str(response.status_code).startswith('2'):
            return True
        raise AcmeError(response)

    def new_authorization(self, domain):
        """
        Requests a new authorization for the specified domain.
        """
        response = self.post('/acme/new-authz', {
            'resource': 'new-authz',
            'identifier': { 'type': 'dns', 'value': domain }
        })
        if response.status_code == 201:
            return NewAuthorizationResult(_json(response), response.headers.get('Location'))
        raise AcmeError(response)

    def validate_authorization(self, uri, _type, key_authorization):
        """
        Marks the specified validation as complete.
        """
        response = self.post(uri, {
            'resource': 'challenge',
            'type': _type,
            'keyAuthorization': key_authorization,
        })
        if str(response.status_code).startswith('2'):
            return True
        raise AcmeError(response)

    def get_authorization(self, uri):
        """
        Returns the authorization status.
        """
        response = self.get(uri)
        try:
            return response.json()
        except (ValueError, TypeError, AttributeError) as e:
            raise AcmeError(e)

    def issue_certificate(self, csr):
        http_headers = { 'Accept': 'application/pkix-cert' }
        response = self.post('/acme/new-cert', {
            'resource': 'new-cert',
            'csr': csr,
        }, headers=http_headers)
        if response.status_code == 201:
            # Get the issuer certificate
            chain = _get_link(response.headers.get('Link'), 'up')
            if chain:
                chain = requests.get(chain, headers=DEFAULT_HEADERS).content

            return IssuanceResult(
                response.content,
                response.headers.get('Location'),
                chain,
            )
        raise AcmeError(response)

    def revoke_certificate(self, cert):
        response = self.post('/acme/revoke-cert', {
            'resource': 'revoke-cert',
            'certificate': cert,
        })
        if response.status_code == 200:
            return True
        raise AcmeError(response)

    def get(self, path, headers=None):
        _headers = DEFAULT_HEADERS.copy()
        if headers:
            _headers.update(headers)
        return requests.get(self.path(path), headers=_headers)

    def post(self, path, body, headers=None):
        _headers = DEFAULT_HEADERS.copy()
        _headers['Content-Type'] = 'application/json'
        if headers:
            _headers.update(headers)

        header, protected = self.get_headers()
        body = sign_request(self.account.key, header, protected, body)

        return requests.post(self.path(path), data=body, headers=_headers)

    def path(self, path):
        # Make sure path is relative
        if path.startswith('http'):
            path = urlparse(path).path
        return urljoin(self.url, path)


RegistrationResult = namedtuple('RegistrationResult', 'contents uri terms')
NewAuthorizationResult = namedtuple('NewAuthorizationResult', 'contents uri')
IssuanceResult = namedtuple('IssuanceResult', 'certificate location intermediate')


def _get_link(header, type):
    links = (header or '').split(',')
    for link in links:
        link = link.strip()
        if type in link:
            return link[1:link.index('>')]

def _json(response):
    try:
        return response.json()
    except ValueError as e:
        raise AcmeError("Invalid JSON response. {}".format(e))



"""
The domain authorization command. Authorizations last up to 300 days on the
production Let's Encrypt service at the time of writing, so it makes sense to
separate certificate issuance from ownership verification.
"""

import logging
import time
import hashlib

from .acme import Acme
from .crypto import generate_jwk_thumbprint, jose_b64
from .errors import ManualeError, AcmeError
from .helpers import confirm

logger = logging.getLogger(__name__)

def authorize(server, account, domains):
    acme = Acme(server, account)
    thumbprint = generate_jwk_thumbprint(account.key)

    try:
        # Get pending authorizations for each domain
        authz = {}
        for domain in domains:
            logger.info("Requesting challenge for {}.".format(domain))
            created = acme.new_authorization(domain)
            auth = created.contents
            auth['uri'] = created.uri

            # Find the DNS challenge
            try:
                auth['challenge'] = [ch for ch in auth.get('challenges', []) if ch.get('type') == 'dns-01'][0]
            except IndexError:
                raise ManualeError("Manuale only supports the dns-01 challenge. The server did not return one.")

            auth['key_authorization'] = "{}.{}".format(auth['challenge'].get('token'), thumbprint)
            digest = hashlib.sha256()
            digest.update(auth['key_authorization'].encode('ascii'))
            auth['txt_record'] = jose_b64(digest.digest())

            authz[domain] = auth

        logger.info("")
        logger.info("DNS verification required. Make sure these records are in place:")
        logger.info("")
        for domain in domains:
            auth = authz[domain]
            logger.info("  _acme-challenge.{}. \"{}\"".format(domain, auth['txt_record']))
        logger.info("")
        input("Press enter to continue.")

        # Verify each domain
        done, failed = set(), set()
        for domain in domains:
            logger.info("")
            auth = authz[domain]
            challenge = auth['challenge']
            acme.validate_authorization(challenge['uri'], 'dns-01', auth['key_authorization'])

            while True:
                logger.info("{}: waiting for verification. Checking in 5 seconds.".format(domain))
                time.sleep(5)

                response = acme.get_authorization(auth['uri'])
                status = response.get('status')
                if status == 'valid':
                    done.add(domain)
                    logger.info("{}: OK! Authorization lasts until {}.".format(domain, response.get('expires', '(not provided)')))
                    break
                elif status != 'pending':
                    failed.add(domain)

                    # Failed, dig up details
                    error_type, error_reason = "unknown", "N/A"
                    try:
                        challenge = [ch for ch in response.get('challenges', []) if ch.get('type') == 'dns-01'][0]
                        error_type = challenge.get('error').get('type')
                        error_reason = challenge.get('error').get('detail')
                    except (ValueError, IndexError, AttributeError, TypeError):
                        pass

                    logger.info("{}: {} ({})".format(domain, error_reason, error_type))
                    break

        logger.info("")
        if failed:
            logger.info("{} domain(s) authorized, {} failed.".format(len(done), len(failed)))
            logger.info("Authorized: {}".format(' '.join(done) or "N/A"))
            logger.info("Failed: {}".format(' '.join(failed)))
        else:
            logger.info("{} domain(s) authorized. Let's Encrypt!".format(len(done)))
    except IOError as e:
        logger.error("A connection or service error occurred. Aborting.")
        raise ManualeError(e)



"""
A small container for account data. Ideally we would get away with just using a
PEM-encoded RSA private key as the account file, but we might need the account
URI at some point.
"""

import json

from .crypto import load_rsa_key, export_rsa_key


class Account:

    def __init__(self, key, uri=None):
        self.key = key
        self.uri = uri

    def serialize(self):
        return json.dumps({
            'key': export_rsa_key(self.key).decode('utf-8'),
            'uri': self.uri,
        }).encode('utf-8')


def deserialize(data):
    try:
        if not isinstance(data, str):
            data = data.decode('utf-8')
        data = json.loads(data)
        if 'key' not in data or 'uri' not in data:
            raise ValueError("Missing 'key' or 'uri' fields.")
        return Account(key=load_rsa_key(data['key'].encode('utf8')), uri=data['uri'])
    except (TypeError, ValueError, AttributeError) as e:
        raise IOError("Invalid account structure: {}".format(e))



"""
The certificate revocation command.
"""

import logging

from .acme import Acme
from .errors import ManualeError
from .crypto import (
    load_pem_certificate,
    export_certificate_for_acme,
    get_certificate_domains
)
from .helpers import confirm

logger = logging.getLogger(__name__)

def revoke(server, account, certificate):
    # Load the certificate
    try:
        with open(certificate, 'rb') as f:
            certificate = load_pem_certificate(f.read())
    except IOError as e:
        logger.error("Couldn't read the certificate.")
        raise ManualeError(e)

    # Confirm
    logger.info("Are you sure you want to revoke this certificate? It includes the following domains:")
    for domain in get_certificate_domains(certificate):
        logger.info("  {}".format(domain))
    if not confirm("This can't be undone. Confirm?", default=False):
        raise ManualeError("Aborting.")

    # Revoke.
    acme = Acme(server, account)
    try:
        acme.revoke_certificate(export_certificate_for_acme(certificate))
    except IOError as e:
        raise ManualeError(e)

    logger.info("Certificate revoked.")



"""
The command line interface.
"""

import argparse
import logging
import sys
import os

from .account import Account
from .account import deserialize as deserialize_account
from .authorize import authorize
from .issue import issue
from .info import info
from .register import register
from .revoke import revoke
from .errors import ManualeError
import manuale

logger = logging.getLogger(__name__)

# Text
DESCRIPTION = \
"""
Interact with ACME certification authorities such as Let's Encrypt.

No idea what you're doing? Register an account, authorize your domains and
issue a certificate or two. Call a command with -h for more instructions.
"""

DESCRIPTION_REGISTER = \
"""
Creates a new account key and registers on the server. The resulting --account
is saved in the specified file, and required for most other operations.

You only have to do this once. Keep the account file safe and secure: it
contains your private key, and you need it to get certificates!
"""

DESCRIPTION_AUTHORIZE = \
"""
Authorizes a domain or multiple domains for your account through DNS
verification. You will need to set DNS records as prompted.

After authorizing a domain, you can issue certificates for it. Authorizations
can last for a long time, so you might not need to do this every time you want
a new certificate.  This depends on the server being used. You should see an
expiration date for the authorization after completion.

ACME does not currently provide a read-only mechanism for determining if a
domain is already authorized. This means that you might go through this even if
you don't need to.
"""

DESCRIPTION_ISSUE = \
"""
Issues a certificate for one or more domains. Hopefully needless to say, you
must have valid authorizations for the domains you specify first.

This will generate a new RSA key and CSR for you. But if you want, you can
bring your own with the --key-file and --csr-file attributes. You can also set
a custom --key-size. (Don't try something stupid like 512, the server won't
accept it. I tried.)

The resulting key and certificate are written into domain.pem and domain.crt.
A chained certificate with the intermediate included is also written to
domain.chain.crt. You can change the --output directory to something else from
the working directory as well.

(If you're passing your own CSR, the given domains can be whatever you want.)

Note that unlike many other certification authorities, ACME does not add a
non-www or www alias to certificates. If you want this to happen, add it
yourself. You need to authorize both as well.

Certificate issuance has a server-side rate limit. Don't overdo it.
"""

DESCRIPTION_REVOKE = \
"""
Revokes a certificate. The certificate must have been issued using the
current account.
"""

DESCRIPTION_INFO = \
"""
Shows raw registration info for the current account.
"""

# Defaults
LETS_ENCRYPT_PRODUCTION = "https://acme-v01.api.letsencrypt.org/"
DEFAULT_ACCOUNT_PATH = 'account.json'
DEFAULT_CERT_KEY_SIZE = 2048

# Command handlers
def _register(args):
    register(args.server, args.account, args.email)

def _authorize(args):
    account = load_account(args.account)
    authorize(args.server, account, args.domain)

def _issue(args):
    account = load_account(args.account)
    issue(
        server=args.server,
        account=account,
        domains=args.domain,
        key_size=args.key_size,
        key_file=args.key_file,
        csr_file=args.csr_file,
        output_path=args.output
    )

def _revoke(args):
    account = load_account(args.account)
    revoke(
        server=args.server,
        account=account,
        certificate=args.certificate
    )

def _info(args):
    account = load_account(args.account)
    info(args.server, account)

def load_account(path):
    # Show a more descriptive message if the file doesn't exist.
    if not os.path.exists(path):
        logger.error("Couldn't find an account file at {}.".format(path))
        logger.error("Are you in the right directory? Did you register yet?")
        logger.error("Run 'manuale -h' for instructions.")
        raise ManualeError()

    try:
        with open(path, 'rb') as f:
            return deserialize_account(f.read())
    except (ValueError, IOError) as e:
        logger.error("Couldn't read account file. Aborting.")
        raise ManualeError(e)

class Formatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    pass

# Where it all begins.
def main():
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        formatter_class=Formatter,
    )
    subparsers = parser.add_subparsers()

    # Server switch
    parser.add_argument('--server', '-s', help="The ACME server to use", default=LETS_ENCRYPT_PRODUCTION)
    parser.add_argument('--account', '-a', help="The account file to use or create", default=DEFAULT_ACCOUNT_PATH)

    # Account creation
    register = subparsers.add_parser(
        'register',
        help="Create a new account and register",
        description=DESCRIPTION_REGISTER,
        formatter_class=Formatter,
    )
    register.add_argument('email', type=str, help="Account e-mail address")
    register.set_defaults(func=_register)

    # Domain verification
    authorize = subparsers.add_parser(
        'authorize',
        help="Verify domain ownership",
        description=DESCRIPTION_AUTHORIZE,
        formatter_class=Formatter,
    )
    authorize.add_argument('domain', help="One or more domain names to authorize", nargs='+')
    authorize.set_defaults(func=_authorize)

    # Certificate issuance
    issue = subparsers.add_parser(
        'issue',
        help="Request a new certificate",
        description=DESCRIPTION_ISSUE,
        formatter_class=Formatter,
    )
    issue.add_argument('domain', help="One or more domain names to include in the certificate", nargs='+')
    issue.add_argument('--key-size', '-b', help="The key size to use for the certificate", type=int, default=DEFAULT_CERT_KEY_SIZE)
    issue.add_argument('--key-file', '-k', help="Existing key file to use for the certificate")
    issue.add_argument('--csr-file', help="Existing signing request to use")
    issue.add_argument('--output', '-o', help="The output directory for created objects", default='.')
    issue.set_defaults(func=_issue)

    # Certificate revocation
    revoke = subparsers.add_parser(
        'revoke',
        help="Revoke an issued certificate",
        description=DESCRIPTION_REVOKE,
        formatter_class=Formatter,
    )
    revoke.add_argument('certificate', help="The certificate file to revoke")
    revoke.set_defaults(func=_revoke)

    # Account info
    info = subparsers.add_parser(
        'info',
        help="Shows account information from the service",
        description=DESCRIPTION_INFO,
        formatter_class=Formatter,
    )
    info.set_defaults(func=_info)

    # Version
    version = subparsers.add_parser('version', help="Show the version number")
    version.set_defaults(func=lambda *args: logger.info("manuale {}".format(manuale.__version__)))

    # Parse
    args = parser.parse_args()
    if not hasattr(args, 'func'):
        parser.print_help()
        sys.exit(0)

    # Set up logging
    root = logging.getLogger('manuale')
    root.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(logging.Formatter("%(message)s"))
    root.addHandler(handler)

    # Let's encrypt
    try:
        args.func(args)
    except ManualeError as e:
        if str(e):
            logger.error(e)
        sys.exit(1)
    except KeyboardInterrupt:
        logger.error("")
        logger.error("Interrupted.")
        sys.exit(2)
    except Exception as e:
        logger.error("Oops! An unhandled error occurred. Please file a bug.")
        logger.exception(e)
        sys.exit(3)



def confirm(msg, default=True):
    while True:
        choices = 'Y/n' if default else 'y/N'
        answer = input("{} [{}] ".format(msg, choices)).strip().lower()

        if answer in { 'yes', 'y' } or (default and not answer):
            return True
        elif answer in { 'no', 'n' } or (not default and not answer):
            return False



from setuptools import setup
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    readme = f.read()

setup(
    name='manuale',
    version='1.0.0',

    license='MIT',
    description="A fully manual Let's Encrypt/ACME client",
    long_description=readme,
    url='https://github.com/veeti/manuale',
    author="Veeti Paananen",
    author_email='veeti.paananen@rojekti.fi',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    packages=['manuale'],
    install_requires=[
        'cryptography >= 1.0',
        'requests',
    ],

    entry_points={
        'console_scripts': [
            'manuale = manuale.cli:main',
        ],
    },
)



"""
The moment you've been waiting for: actually getting SSL. For free!
"""

import binascii
import os
import logging

from cryptography.hazmat.primitives.hashes import SHA256

from .acme import Acme
from .errors import ManualeError
from .crypto import (
    generate_rsa_key,
    load_rsa_key,
    export_rsa_key,
    create_csr,
    load_csr,
    export_csr_for_acme,
    load_der_certificate,
    export_pem_certificate,
)
from .helpers import confirm

logger = logging.getLogger(__name__)

EXPIRATION_FORMAT = "%Y-%m-%d"

def issue(server, account, domains, key_size, key_file=None, csr_file=None, output_path=None):
    if not output_path or output_path == '.':
        output_path = os.getcwd()

    # Load key or generate
    if key_file:
        try:
            with open(key_file, 'rb') as f:
                certificate_key = load_rsa_key(f.read())
        except (ValueError, AttributeError, TypeError, IOError) as e:
            logger.error("Couldn't read certificate key.")
            raise ManualeError(e)
    else:
        logger.info("Generating a {} bit RSA key. This might take a second.".format(key_size))
        certificate_key = generate_rsa_key(key_size)
        logger.info("Key generated.")
        logger.info("")

    # Load CSR or generate
    if csr_file:
        try:
            with open(csr_file, 'rb') as f:
                csr = export_csr_for_acme(load_csr(f.read()))
        except (ValueError, AttributeError, TypeError, IOError) as e:
            logger.error("Couldn't read CSR.")
            raise ManualeError(e)
    else:
        csr = create_csr(certificate_key, domains)

    acme = Acme(server, account)
    try:
        logger.info("Requesting certificate issuance...")
        result = acme.issue_certificate(csr)
        logger.info("Certificate issued.")
    except IOError as e:
        logger.error("Connection or service request failed. Aborting.")
        raise ManualeError(e)

    try:
        certificate = load_der_certificate(result.certificate)

        # Print some neat info
        logger.info("")
        logger.info("  Expires: {}".format(certificate.not_valid_after.strftime(EXPIRATION_FORMAT)))
        logger.info("   SHA256: {}".format(binascii.hexlify(certificate.fingerprint(SHA256())).decode('ascii')))
        logger.info("")

        # Write the key, certificate and full chain
        os.makedirs(output_path, exist_ok=True)
        cert_path = os.path.join(output_path, domains[0] + '.crt')
        chain_path = os.path.join(output_path, domains[0] + '.chain.crt')
        intermediate_path = os.path.join(output_path, domains[0] + '.intermediate.crt')
        key_path = os.path.join(output_path, domains[0] + '.pem')

        with open(key_path, 'wb') as f:
            f.write(export_rsa_key(certificate_key))
            logger.info("Wrote key to {}".format(f.name))

        with open(cert_path, 'wb') as f:
            f.write(export_pem_certificate(certificate))
            logger.info("Wrote certificate to {}".format(f.name))

        with open(chain_path, 'wb') as f:
            f.write(export_pem_certificate(certificate))
            if result.intermediate:
                f.write(export_pem_certificate(load_der_certificate(result.intermediate)))
            logger.info("Wrote certificate with intermediate to {}".format(f.name))

        if result.intermediate:
            with open(intermediate_path, 'wb') as f:
                f.write(export_pem_certificate(load_der_certificate(result.intermediate)))
                logger.info("Wrote intermediate certificate to {}".format(f.name))
    except IOError as e:
        logger.error("Failed to write certificate or key. Going to print them for you instead.")
        logger.error("")
        for line in export_rsa_key(certificate_key).decode('ascii').split('\n'):
            logger.error(line)
        for line in export_pem_certificate(certificate).decode('ascii').split('\n'):
            logger.error(line)
        raise ManualeError(e)



class ManualeError(Exception):
    pass


class AcmeError(IOError):
    def __init__(self, response):
        message = "The ACME request failed."
        try:
            details = response.json()
            self.type = details.get('type', 'unknown')
            message = "{} (type {}, HTTP {})".format(details.get('detail'),
                    self.type, response.status_code)
        except (ValueError, TypeError, AttributeError):
            pass
        super().__init__(message)


class AccountAlreadyExistsError(AcmeError):

    def __init__(self, response, existing_uri):
        super().__init__(response)
        self.existing_uri = existing_uri



"""
Cryptography, hopefully mostly correct.
"""

import base64
import json
import logging

from cryptography import x509
from cryptography.x509 import NameOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric.rsa import (
    generate_private_key,
    RSAPrivateKey,
)
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    Encoding,
    PrivateFormat,
    NoEncryption,
)
from cryptography.hazmat.primitives import hashes

logger = logging.getLogger(__name__)

def jose_b64(data):
    """
    Encodes data with JOSE/JWS base 64 encoding.
    """
    return base64.urlsafe_b64encode(data).decode('ascii').replace('=', '')

def generate_rsa_key(size=2048):
    """
    Generates a new RSA private key.
    """
    return generate_private_key(65537, size, default_backend())

def generate_header(account_key):
    """
    Creates a new request header for the specified account key.
    """
    numbers = account_key.public_key().public_numbers()
    e = numbers.e.to_bytes((numbers.e.bit_length() // 8 + 1), byteorder='big')
    n = numbers.n.to_bytes((numbers.n.bit_length() // 8 + 1), byteorder='big')
    if n[0] == 0: # for strict JWK
        n = n[1:]
    return {
        'alg': 'RS256',
        'jwk': {
            'kty': 'RSA',
            'e': jose_b64(e),
            'n': jose_b64(n),
        },
    }

def generate_jwk_thumbprint(account_key):
    """
    Generates a JWK thumbprint for the specified account key.
    """
    jwk = generate_header(account_key)['jwk']
    as_json = json.dumps(jwk, sort_keys=True, separators=(',', ':'))

    sha256 = hashes.Hash(hashes.SHA256(), default_backend())
    sha256.update(as_json.encode('utf-8'))

    return jose_b64(sha256.finalize())

def sign_request(key, header, protected_header, payload):
    """
    Creates a JSON Web Signature for the request header and payload using the
    specified account key.
    """
    protected = jose_b64(json.dumps(protected_header).encode('utf8'))
    payload = jose_b64(json.dumps(payload).encode('utf8'))

    signer = key.signer(padding.PKCS1v15(), hashes.SHA256())
    signer.update(protected.encode('ascii'))
    signer.update(b'.')
    signer.update(payload.encode('ascii'))

    return json.dumps({
        'header': header,
        'protected': protected,
        'payload': payload,
        'signature': jose_b64(signer.finalize()),
    })

def load_rsa_key(data):
    """
    Loads a PEM-encoded RSA private key.
    """
    key = load_pem_private_key(data, password=None, backend=default_backend())
    if not isinstance(key, RSAPrivateKey):
        raise ValueError("Key is not a private RSA key.")
    elif key.key_size < 2048:
        raise ValueError("The key must be 2048 bits or longer.")
    return key

def export_rsa_key(key):
    """
    Exports a private RSA key in OpenSSL PEM format.
    """
    return key.private_bytes(Encoding.PEM, PrivateFormat.TraditionalOpenSSL, NoEncryption())

def create_csr(key, domains):
    """
    Creates a CSR in DER format for the specified key and domain names.
    """
    assert domains
    name = x509.Name([
        x509.NameAttribute(NameOID.COMMON_NAME, domains[0]),
    ])
    san = x509.SubjectAlternativeName([x509.DNSName(domain) for domain in domains])
    csr = x509.CertificateSigningRequestBuilder().subject_name(name) \
        .add_extension(san, critical=False) \
        .sign(key, hashes.SHA256(), default_backend())
    return export_csr_for_acme(csr)

def export_csr_for_acme(csr):
    """
    Exports a X.509 CSR for the ACME protocol (JOSE Base64 DER).
    """
    return jose_b64(csr.public_bytes(Encoding.DER))

def load_csr(data):
    """
    Loads a PEM X.509 CSR.
    """
    return x509.load_pem_x509_csr(data, default_backend())

def load_der_certificate(data):
    """
    Loads a DER X.509 certificate.
    """
    return x509.load_der_x509_certificate(data, default_backend())

def load_pem_certificate(data):
    """
    Loads a PEM X.509 certificate.
    """
    return x509.load_pem_x509_certificate(data, default_backend())

def get_certificate_domains(cert):
    """
    Gets a list of all Subject Alternative Names in the specified certificate.
    """
    for ext in cert.extensions:
        ext = ext.value
        if isinstance(ext, x509.SubjectAlternativeName):
            return ext.get_values_for_type(x509.DNSName)
    return []

def export_pem_certificate(cert):
    """
    Exports a X.509 certificate as PEM.
    """
    return cert.public_bytes(Encoding.PEM)

def export_certificate_for_acme(cert):
    """
    Exports a X.509 certificate for the ACME protocol (JOSE Base64 DER).
    """
    return jose_b64(cert.public_bytes(Encoding.DER))



"""
The account info command.
"""

import json
import logging
import sys

from .acme import Acme
from .errors import ManualeError

logger = logging.getLogger(__name__)

def info(server, account):
    acme = Acme(server, account)

    try:
        logger.info("Requesting account data...")
        reg = acme.get_registration()
        sys.stdout.write(json.dumps(reg, indent=4, sort_keys=True))
        sys.stdout.flush()
    except IOError as e:
        raise ManualeError(e)
